#include "config.h"

#include <stdio.h>
#include <stdlib.h>

#ifdef HAVE_MALLOC_H
#include <malloc.h>
#endif

#include "ga_dyads.h"
#include "defines.h"
#include "random.h"

void crossover(Chrs **dyad,int numWordGroup,Words *word,int minSpaceWidth,int maxSpaceWidth,
   Wheel *wheel,int populationSize,Fitness *fitness,char *motifUniq) {

   register int i,j;
   int found,id1,id2,popuCn,kmerGroup,numIdentical;
   Chrs **tmpDyad;
   double rand,du;

   tmpDyad=alloc_chrs(populationSize,4);

   // only save non-identical top-ranked spaced dyads: no crossover
   popuCn=0;
   for (i=0; i<populationSize; i++) {
      if (motifUniq[i]=='1') {
         for (j=0; j<3; j++) {
            tmpDyad[popuCn][j].wordID   =dyad[fitness[i].index][j].wordID;
            tmpDyad[popuCn][j].wordGroup=dyad[fitness[i].index][j].wordGroup;
         }
         popuCn++;
      }
   }
   // printf("popuCn: %d\n",popuCn);

   // crossover
   numIdentical=0;
   while (popuCn<(int)(0.95*populationSize)) {
      do {
         // pick the first "chromosome"
         du=(double)populationSize*genrand();
         found=0; id1=0; id2=0;
         for (j=0; j<populationSize; j++) {
            if (du>=wheel[j].start && du<=wheel[j].end) { id1=wheel[j].index; found=1; break; }
         }
         if (!found) {
            id1=(int)(populationSize*genrand());
            if (id1==populationSize) id1--;
         }
         // pick the second "chromosome"
         du=(double)populationSize*genrand();
         found=0;
         for (j=0; j<populationSize; j++) {
            if (du>=wheel[j].start && du<=wheel[j].end) { id2=wheel[j].index; found=1; break; }
         }
         if (!found) {
            id2=(int)(populationSize*genrand());
            if (id2==populationSize) id2--;
         }
         // printf("id1 id2: %d %d\n",id1,id2);
      } while (id1==id2); // do not pick the same "chromosome"
  
      rand=genrand();
      if (rand>=0 && rand<1/3) {
         // replace w1 of dyad1 by w1 of dyad2
         tmpDyad[popuCn][0].wordID   =dyad[id2][0].wordID;       // replacement
         tmpDyad[popuCn][1].wordID   =dyad[id1][1].wordID; 
         tmpDyad[popuCn][2].wordID   =dyad[id1][2].wordID;
         tmpDyad[popuCn][0].wordGroup=dyad[id2][0].wordGroup;    // replacement
         tmpDyad[popuCn][1].wordGroup=dyad[id1][1].wordGroup; 
         tmpDyad[popuCn][2].wordGroup=dyad[id1][2].wordGroup;
         popuCn++;
         if (popuCn==populationSize) break;
         
         // replace w1 of dyad2 by w1 of dyad1
         tmpDyad[popuCn][0].wordID   =dyad[id1][0].wordID;       // replacement
         tmpDyad[popuCn][1].wordID   =dyad[id2][1].wordID; 
         tmpDyad[popuCn][2].wordID   =dyad[id2][2].wordID;
         tmpDyad[popuCn][0].wordGroup=dyad[id1][0].wordGroup;    // replacement
         tmpDyad[popuCn][1].wordGroup=dyad[id2][1].wordGroup; 
         tmpDyad[popuCn][2].wordGroup=dyad[id2][2].wordGroup;
         popuCn++;
         if (popuCn==populationSize) break;
      }
      else if (rand>=1/3 && rand<2/3) {
         // replace spacer of dyad1 by spacer of dyad2
         tmpDyad[popuCn][0].wordID   =dyad[id1][0].wordID;
         tmpDyad[popuCn][1].wordID   =dyad[id2][1].wordID;       // replacement
         tmpDyad[popuCn][2].wordID   =dyad[id1][2].wordID;
         tmpDyad[popuCn][0].wordGroup=dyad[id1][0].wordGroup;
         tmpDyad[popuCn][1].wordGroup=dyad[id2][1].wordGroup;    // replacement 
         tmpDyad[popuCn][2].wordGroup=dyad[id1][2].wordGroup;
         popuCn++;
         if (popuCn==populationSize) break;

         // replace spacer dyad2 by spacer of dyad1
         tmpDyad[popuCn][0].wordID   =dyad[id2][0].wordID;
         tmpDyad[popuCn][1].wordID   =dyad[id1][1].wordID;       // replacement
         tmpDyad[popuCn][2].wordID   =dyad[id2][2].wordID;      
         tmpDyad[popuCn][0].wordGroup=dyad[id2][0].wordGroup;
         tmpDyad[popuCn][1].wordGroup=dyad[id1][1].wordGroup;    // replacement 
         tmpDyad[popuCn][2].wordGroup=dyad[id2][2].wordGroup;
         popuCn++;
         if (popuCn==populationSize) break;
      }
      else {
         // replace w2 of dyad1 by w2 of dyad2
         tmpDyad[popuCn][0].wordID   =dyad[id1][0].wordID;       
         tmpDyad[popuCn][1].wordID   =dyad[id1][1].wordID; 
         tmpDyad[popuCn][2].wordID   =dyad[id2][2].wordID;       // replacement
         tmpDyad[popuCn][0].wordGroup=dyad[id1][0].wordGroup;   
         tmpDyad[popuCn][1].wordGroup=dyad[id1][1].wordGroup; 
         tmpDyad[popuCn][2].wordGroup=dyad[id2][2].wordGroup;    // replacement
         popuCn++;
         if (popuCn==populationSize) break;

         // replace w2 of dyad2 by w2 of dyad1
         tmpDyad[popuCn][0].wordID   =dyad[id2][0].wordID;       
         tmpDyad[popuCn][1].wordID   =dyad[id2][1].wordID; 
         tmpDyad[popuCn][2].wordID   =dyad[id1][2].wordID;       // replacement
         tmpDyad[popuCn][0].wordGroup=dyad[id2][0].wordGroup;
         tmpDyad[popuCn][1].wordGroup=dyad[id2][1].wordGroup; 
         tmpDyad[popuCn][2].wordGroup=dyad[id1][2].wordGroup;    // replacement
         popuCn++;
         if (popuCn==populationSize) break;
      }
      // printf("popuCn: %d\n",popuCn);
   }

   // introduce newly generated spaced dyads
   while (popuCn<populationSize) {
      // initialize w1 of dyad: w1-spacer-w2 
      kmerGroup=(int)(numWordGroup*genrand());
      if (kmerGroup==numWordGroup) kmerGroup--;
      tmpDyad[popuCn][0].wordGroup=kmerGroup;

      rand=genrand();
      tmpDyad[popuCn][0].wordID=0;  // update right below
      for (j=0; j<word[kmerGroup].count; j++) {
         if (rand>=word[kmerGroup].prob_sta[j] && rand<word[kmerGroup].prob_end[j]) { tmpDyad[popuCn][0].wordID=j; break; }
      }

      // choose a spacer wIDth
      tmpDyad[popuCn][1].wordGroup=-1;  // dummy
      tmpDyad[popuCn][1].wordID=minSpaceWidth+(int)((maxSpaceWidth-minSpaceWidth+1)*genrand());

      // initialize w2 of dyad w1-spacer-w2 
      kmerGroup=(int)(numWordGroup*genrand());
      if (kmerGroup==numWordGroup) kmerGroup--;
      tmpDyad[popuCn][2].wordGroup=kmerGroup;

      tmpDyad[popuCn][2].wordID=0; 
      rand=genrand();
      for (j=0; j<word[kmerGroup].count; j++) {
         if (rand>=word[kmerGroup].prob_sta[j] && rand<word[kmerGroup].prob_end[j]) { tmpDyad[popuCn][2].wordID=j; break; }
      }
      popuCn++;
   }


   // update the population - both pwm lengths and members
   for (i=0; i<populationSize; i++) {
      for (j=0; j<3; j++) {
         dyad[i][j].wordID   =tmpDyad[i][j].wordID;
         dyad[i][j].wordGroup=tmpDyad[i][j].wordGroup;
      }
   }

   if (tmpDyad[0]) { free(tmpDyad[0]);  tmpDyad[0]=NULL; }
   if (tmpDyad)    { free(tmpDyad);     tmpDyad=NULL;    }
}

