#include <stdlib.h>
#include <string.h>
#include <stdio.h>

typedef struct ktuples_info Ktuples;
struct ktuples_info {
   char *seq;
   double z,eCn,stdCn,p;
   int count;
};

typedef struct word_info Words;
struct word_info {
   char **s1;
   double *prob_sta,*prob_end;
   int count;
};

typedef struct chromosome_info Chrs;
struct chromosome_info {
   int wordID,wordGroup;
};

typedef struct site_info Sites;
struct site_info {
   int    seq,pos;
   char   rev;
   double pvalue;
};

typedef struct fitness_score Fitness;
struct fitness_score {
   double value;
   int    index;
};

typedef struct roulette Wheel;
struct roulette {
   double start,end;
   int index;
   // char  I_evalue;
};

//pwm score distribution
typedef struct M_function Mfunction;
struct M_function{
     // int col;
     int score;
     double pvalue;
};

int *alloc_int (int );
int **alloc_int_int(int ,int );
int ***alloc_int_int_int(int ,int ,int );
int Compare_fitness(const void *, const void *);
int Compare_double(const void *, const void *);
int Compare_M1(const void *s1, const void *s2);
int Compare_z(const void *s1, const void *s2);
int Compare_site(const void *s1, const void *s2);
int pwm_score_dist(double *,int **,int ,Mfunction *);
int ini_M(Mfunction *,double *,int **,int );
int prod_M(Mfunction *,int ,Mfunction *,int ,Mfunction *);
int scan_sequence(Sites *,int ,char **,char **,int *,int **,int ,int ,char *,double *);
int scan_sequence_full(Mfunction *,int ,Sites *,int ,char **,char **,int *,int **,int ,int ,double *);
int determine_cutoff(Mfunction *,int ,double );
int check_pwm_uniqueness_dist(double ***,int *,int ,Fitness *,double ,double ,char *,int );
int top_kmer(Words *,Ktuples *,int *,int );

char *alloc_char(int );
char **alloc_char_char(int ,int );
char **read_seq(int *,int *,char **,int ,int ,double *,char *);

void sgenrand(unsigned long long );
void reverse_seq(char **,char **,int ,int *);
void roulett_wheel_weight(Fitness *,int ,Wheel *);
void roulett_wheel_fitness(Fitness *,int ,Wheel *);
void sort_fitness(Fitness *,int );
void sort_double(double *,int );
void sort_ptable(Mfunction *M1,int size);
void sort_kmer_z(Ktuples *,int );
void sort_sites(Sites *,int );
void sort_sites(Sites *,int );
void observed_pwm(Sites *,int ,char **,char **,int ,int ,double **);
void mask_sites(int ,char **,char **,int *,Sites *,int );
void extend_alignment(Sites *,int ,char **,char **,int *,int ,int ,int *);
void align_sites_count(Sites *,char **,char **,int ,int ,double **) ;
void initialisation(Chrs **,int ,int ,Words *,int ,int );
void log_pwm(double **,double **,int );
void log_odds_to_int(double **,int **,int ,double *);
void pwm_double_to_int(double **,int **,int );
void construct_pwm(double **,double **,double **,char **,char **,int *,int ,int ,char *);
void score_subsequences(int ,char **,char **,int *,double **,int ,double **,double **,char *,double *);
void normalize_global(double **,double **,int *,int ,int ,char *,int );
void copy_pwm_int_to_double(double **,double **,int );
void copy_pwm(double **,double **,int );
void dyad_to_pwm(Words *,int ,Chrs **,double ***,int *);
void destroy_word(Words *word,int );
void destroy_ktuples(Ktuples *,int );
void crossover(Chrs **,int ,Words *,int ,int ,Wheel *,int ,Fitness *,char *);
void mutation(Chrs **,int ,Words *,int ,int ,Wheel *,int ,Fitness *,char *);
void count_k_tuples(char **,char **,int ,int *,char **,int ,int ,int *);
void enumerate_kmers(char **,char **,char **,char **);
int word_for_dyad(Words *,char **,char **,int ,int *,double *,int *,int *,int *,int *);
void sample_without_replacement(char *,int ,int );
void print_result_2(Sites *,int ,int ,char **,char **,int *,char **,double ,double **,int ,int ,char *,char *,int ,double ,FILE *,FILE *);
void print_motif(Sites *,int ,char **,char **,int *,int ,int );
void pwm_profile(double **,int ,char *);
void consensus_pwm(double **,int ,char *);
void effect_seq_length(char **,int ,int *,char *,int *);
void effect_seq_length_full(char **,int ,int *,int *);
void reassign_fitness(int ,char *,int ,Fitness *);
void standardize_pwm(double **,int );
void score_kmers(Ktuples *,double *,int ,int *,char **,int );

double *alloc_double(int );
double **alloc_double_double(int ,int );
double ***alloc_double_double_double(int ,int ,int );
double alignment(Sites *,int ,char **,char **,int ,int );
double meme_e_value(double *,double **,int ,int ,int ,int *);
double p_value(int ,Mfunction *,int );
double E_value(double **, int , double *, int , int , int *);
double *base_frequency(int ,char **,int *);
double vector_similarity(void);
double check_convergence(double **,double **,int );
double find_pvalue(int ,Mfunction *,int );

Chrs **alloc_chrs(int ,int );
Words *alloc_word(int ,int );
Sites *alloc_site(int );
Sites **alloc_site_site(int size1,int size2);
Wheel *alloc_wheel(int );
Fitness *alloc_fitness(int );
Ktuples *alloc_ktuples(int ,int );

