#include "config.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#ifdef HAVE_MALLOC_H
#include <malloc.h>
#endif

#include <math.h>
#include "defines.h"
#include "ga_dyads.h"
#include "random.h"

void initialisation(Chrs **dyad,int populationSize,int numWordGroup,Words *word,
   int minSpaceWidth,int maxSpaceWidth) {

   register int j;
   int group,popuCn;
   double rand;

   //for (j=0; j<word[0].count; j++) printf("%s\n",word[0].s1[j]);

   // a population of "chromosomes", each consists of two words separated by a spacer
   popuCn=0; 
   do { 
      // initialize w1 of dyad: w1-spacer-w2 
      group=(int)(numWordGroup*genrand());
      if (group==numWordGroup) group--;
      dyad[popuCn][0].wordGroup=group;

      rand=genrand();
      dyad[popuCn][0].wordID=0;  // update right below
      for (j=0; j<word[group].count; j++) {
         if (rand>=word[group].prob_sta[j] && rand<word[group].prob_end[j]) { 
            dyad[popuCn][0].wordID=j; break; 
         }
      }

      // choose a spacer width
      dyad[popuCn][1].wordGroup=-1;  // dummy
      dyad[popuCn][1].wordID=minSpaceWidth+(int)((maxSpaceWidth-minSpaceWidth+1)*genrand());

      // initialize w2 of dyad w1-spacer-w2 
      group=(int)(numWordGroup*genrand());
      if (group==numWordGroup) group--;
      dyad[popuCn][2].wordGroup=group;

      dyad[popuCn][2].wordID=0; 
      rand=genrand();
      for (j=0; j<word[group].count; j++) {
         if (rand>=word[group].prob_sta[j] && rand<word[group].prob_end[j]) { 
            dyad[popuCn][2].wordID=j; break; 
         }
      }
      popuCn++;
   } while (popuCn<populationSize);
}

void dyad_to_pwm(Words *word,int populationSize,Chrs **dyad,double ***pwm,int *pwmLen) {

   register int i,j,k;
   int len0,len1,len2;
   double major,minor;

   minor=MINOR;
   major=MAJOR;
   for (i=0; i<populationSize; i++) {
      for (j=0; j<MAX_PWM_LENGTH; j++) {
         for (k=0; k<4; k++) pwm[i][j][k]=minor; 
      } 
   }

   for (i=0; i<populationSize; i++) {
      //word 1
      len0=strlen(word[dyad[i][0].wordGroup].s1[dyad[i][0].wordID]); 
      for (j=0; j<len0; j++) {
         switch (word[dyad[i][0].wordGroup].s1[dyad[i][0].wordID][j]) {
            case 'a': pwm[i][j][0]=major; break; 
            case 'c': pwm[i][j][1]=major; break; 
            case 'g': pwm[i][j][2]=major; break; 
            case 't': pwm[i][j][3]=major; break; 
            case 'n': { pwm[i][j][0]=1; pwm[i][j][1]=1; pwm[i][j][2]=1; pwm[i][j][3]=1; break; }
            default: break;
         } 
      }

      //spacer 
      len1=dyad[i][1].wordID;
      for (j=0; j<len1; j++) {
         for (k=0; k<4; k++) pwm[i][len0+j][k]=1; 
      }

      //word 2
      len2=strlen(word[dyad[i][2].wordGroup].s1[dyad[i][2].wordID]);
      for (j=0; j<len2; j++) {
         switch (word[dyad[i][2].wordGroup].s1[dyad[i][2].wordID][j]) {
            case 'a': pwm[i][j+len0+len1][0]=major; break; 
            case 'c': pwm[i][j+len0+len1][1]=major; break; 
            case 'g': pwm[i][j+len0+len1][2]=major; break; 
            case 't': pwm[i][j+len0+len1][3]=major; break; 
            case 'n': { pwm[i][j+len0+len1][0]=1; pwm[i][j+len0+len1][1]=1; pwm[i][j+len0+len1][2]=1; pwm[i][j+len0+len1][3]=1; break; }
            default: break;
         } 
      }
      pwmLen[i]=len0+len1+len2;
   }

   /*---------------------------------------------------------------------------
   for (i=0; i<populationSize; i++) {
      for (k=0; k<4; k++) {
         for (j=0; j<pwmLen[i]; j++) printf("%4.2f ",pwm[i][j][k]); printf("\n");
      } printf("\n");
   }
   -----------------------------------------------------------------------------*/

   /*-------------------------------for testing----------------------------------
   //ggagggnnnnnnnnnncctgg
   pwmLen[0]=21;
  
   for (i=0; i<pwmLen[0]; i++) { for (j=0; j<4; j++) pwm[0][i][j]=0; }

   pwm[0][0][2]=1;
   pwm[0][1][2]=1; 
   pwm[0][2][0]=1;  
   pwm[0][3][2]=1; 
   pwm[0][4][2]=1; 
   pwm[0][5][2]=1; 
   for (i=6; i<16; i++) { for (j=0; j<4; j++) pwm[0][i][j]=1; }
   pwm[0][16][1]=1;
   pwm[0][17][1]=1;
   pwm[0][18][3]=1;
   pwm[0][19][2]=1;
   pwm[0][20][2]=1;

   for (j=0; j<4; j++) {
      for (i=0; i<pwmLen[0]; i++) printf("%2d",pwm[0][i][j]); printf("\n");
   } 
   -------------------------------------------------------------------------*/

#ifdef DEBUG
   for (i=0; i<populationSize; i++) {
      for (k=0; k<4; k++) {
         for (j=0; j<pwmLen[i]; j++) printf("%2d",pwm[i][j][k]); printf("\n");
      } 
   }
#endif
}

