#include "config.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#ifdef HAVE_MALLOC_H
#include <malloc.h>
#endif

#include <math.h>
#include <time.h>
#include <ctype.h>
#include "ga_dyads.h"
#include "defines.h"
#include "random.h"
#include "evalue_meme.h"

int main(int argc,char **argv) {

   register int jjj,ii,jj,i,j;

   // basic settings/info
   int numSeq,maxSeqLen,*seqLen;          // sequence info
   double aveSeqLen;                      // sequence info
   char **seq,**rseq,**geneID;            // sequence info
   char **oseq,**orseq;                   // copy of the original sequences
   double *bfreq;                         // base frequencies
   double *ChIPScore;                     // chip score
   int numSeqWithQualityScore;            // count the number of seqs with quality scores

   // pwms
   double ***pwm;                         // initial population of PWMs from spaced dyads
   int *pwmLen;                           // initial pwm lengths 
   double **t1pwm,**t2pwm;                // two pwms before and after EM steps
   double **opwm2;                        // EM-derived PWM 
   double **logpwm;                       // log-transformed EM-derived EM
   int ***ipwm;                           // integer pwm for computing score distribution
   double ***opwm;                        // observed PWMs from identified sites
   int *pwmnewLen;                        // final motif length after extending to both ends
   Mfunction *scoreDist;
   int pDimension;

   // EM, motif, sites
   double pvalueCutoff;                   // user input, used to determine score cutoff based on ipwm
   int *scoreCutoff;                      // pwm score cutoff for the corresponding p-value cutoff
   double **score,**rscore;               // subsequence score, plus and minus strands
   double logev;                          // log of E-value of a motif;
   int useChIPscore;                      // indicator for using ChIP-seq score for seq. selection for EM
   double ChIPscoreCutoff; 
   int numEM;                             // number of EM steps
   double E_valueCutoff;                  // log E-value cutoff
   int nsitesEM;                          // number of binding sites in sequences subjected to EM
   int minsitesEM;                        // minimal number of sites in a motif in EM sequences
   Sites *siteEM;                         // binding sites in EM sequences
   int *nsites;                           // number of binding sites in full data
   int minsites;                          // minimal number of sites in a motif in full data
   Sites **site;                          // binding sites in all sequences
   int numBasePerSite;                    // miminal number of bases for one binding site
   int motifCn;                           // number of motifs sought and found
   int extTrim;
   int noMotifFound;                      // none of the dyads in the population resulted in a motif
   char **pwmConsensus;                   // consensus sequences of motifs
   double pwmDistCutoff;                  // test statistic for motif pwm similarity
   char *motifUniq;                       // motifs in a population unique or not
   int numUniq;                           // number of unique motifs in a population
   int window;                            // window size for check uniquness
 
   // words for spaced dyad
   Words *word;                           // top-ranked k-mers as the words for spaced dyads
   int numTop3mer,numTop4mer,numTop5mer,numTop6mer;  // No. of top-ranked k-mers as words for dyads
   int maxWordSize;                       // max of the above three
   int numWordGroup;                      // number of non-zero k-mer groups
   int minSpaceWidth,maxSpaceWidth;       // min and max width of spacer of the spaced dyads
   Chrs **dyad;                           // initial population of "chromosomes"
   char **sdyad;                          // char of spaced dyads

   // GA
   int populationSize,numGeneration;      // GA parameters
   Fitness *fitness;                      // "chromosome" fitness
   Wheel *wheel;                          // roulette-wheel selection

   // to speed up only select a subset of sequences for EM algorithm
   double percentSeqEM;                   // percentage of sequences used in EM algorithm
   int numSeqEM,numSeqEMtmp1,numSeqEMtmp2;// number of sequences subject to EM
   char *Iseq;                            // Indicator if a sequence is used in EM or not
   int *emSeqLen;                         // length of sequences used in EM
   double pwmDiff;                        // pwm convergence
   int maxp;                              // initial setting for all motifs maxp=numSeqEM 

   int numCycle;                          // number of GADEM cycles
   int generationNoMotif;                 // maximal number of GA generations in a gadem cycle resulted in no motifs

   // mis.
   seed_t  seed;              // random seed
   int goodArgument,motifCn2,id,numCycleNoMotif,verbose,useAlternativeBackg;
   double sum;
   double *tmpScore;                     // store sorted ChIP-seq quality score
   char *mFileName,*oFileName,*pwmFileName;
   FILE *fp,*fq,*fpwm;
   time_t t;

   printf("\n*-------------------------------------------------------------------------\n");
   printf("|                                                                        |\n");
   printf("|  GADEM: a de novo motif discovery tool for large scale sequence data   |\n");
   printf("|                                                                        |\n");
   printf("|                   Multiple runs are recommended.                       |\n");
   printf("|       Each run GADEM automatically uses a different random seed        |\n");
   printf("*-------------------------------------------------------------------------\n");

   if (argc<3) {

      printf("\n\nUsage: gadem -fseq seqFile [optional arguments]\n\n");
      printf("Optional arguments:\n\n");
      printf("  -gen        maxGeneration   number of GA generations (default: 5)\n");
      printf("  -pop        populationSize  GA population size (default: 100)\n");
      printf("  -maxw3      numTop3mer      number of top-ranked trimers   for dyads (default: 20)\n");
      printf("  -maxw4      numTop4mer      number of top-ranked tetramers for dyads (default: 40)\n");
      printf("  -maxw5      numTop5mer      number of top-ranked pentamers for dyads (default: 60)\n");
      printf("  -maxw6      numTop6mer      number of top-ranked hexamers  for dyads (default: 100)\n");
      printf("  -mingap     minSpacerWidth  minimal no. of unspecifed nucleotides in dyads (default: 0)\n");
      printf("  -maxgap     maxSpacerWidth  maximal no. of unspecifed nucleotides in dyads (default: 10)\n");
      printf("  -score      useScore        indicator for using top-scoring sequences for deriving PWMs\n");
      printf("                              ChIP-seq scores may be stored in sequence header\n");
      printf("                              0 (default) - no (randomly select seqs.), 1 - yes\n");
      printf("  -em         numEM           number of EM steps (default: 40)\n");
      printf("  -fracEM     percentSeqEM    fraction of sequences for EM algorithm (default: 0.5)\n");
      printf("  -fbackg     background      file containing 0th-Markov background (default: estimated from input)\n");
      printf("  -pv         pvalue          pvalue cutoff for declaring binding sites (default: 0.000025)\n");
      printf("  -ev         E_valueCutoff   E-value cutoff for declaring motifs (default: 0)\n");
      printf("  -minN       minsites        minimal number of sites in a motif (default: #base/5000\n");
      printf("  -nbps       numBasePerSite  minimal number of bases per binding site (default: 5000)\n");
      printf("  -extTrim    extTrim         base extension and trimming [default: 1 (1 -yes,0 -no)]\n");
      printf("  -window     window          width of sliding window for comparing motif similarity (default: 6)\n");
      printf("  -fpwm       output          name of output PWM file\n");
      printf("  -fout       output          name of output motif file\n");
      printf("  -verbose    verbose         print on screen [1-yes, 0-no (default)]\n\n");
      exit(0);
   }

   seq=NULL; aveSeqLen=0; maxSeqLen=0; maxSeqLen=MAX_SEQ_LENGTH;
   mFileName=alloc_char(500); mFileName[0]='\0';
   oFileName=alloc_char(500); oFileName[0]='\0';
   pwmFileName=alloc_char(500); pwmFileName[0]='\0';
   seqLen=alloc_int(MAX_NUM_SEQ); geneID=alloc_char_char(MAX_NUM_SEQ,500);
   ChIPScore=alloc_double(MAX_NUM_SEQ);
   numSeq=0; minsites=-1; verbose=0;

   // default settings
   numWordGroup=4;
   numTop3mer=20; numTop4mer=40; numTop5mer=60; numTop6mer=100;
   numGeneration=5; populationSize=100;
   pvalueCutoff=0.000025;
   minSpaceWidth=0; maxSpaceWidth=10;
   useChIPscore=0; 
   numEM=40; percentSeqEM=0.5;
   numBasePerSite=5000;
   useAlternativeBackg=0;
   window=6; numUniq=populationSize;
   E_valueCutoff=0.0; extTrim=1;
   strcpy(oFileName,"gadem.txt"); 
   strcpy(pwmFileName,"observedPWMs.txt"); 
 
   for (ii=1; ii<argc; ii++) {
      if (argv[ii][0]=='-' && isalpha(argv[ii][1])) {
         goodArgument=0;
         if (
             (strncmp(argv[ii],"-ev",3)==0      && strlen(argv[ii])==3) ||
             (strncmp(argv[ii],"-pv",3)==0      && strlen(argv[ii])==3) ||
             (strncmp(argv[ii],"-em",3)==0      && strlen(argv[ii])==3) ||
             (strncmp(argv[ii],"-pop",4)==0     && strlen(argv[ii])==4) ||
             (strncmp(argv[ii],"-gen",4)==0     && strlen(argv[ii])==4) ||
             (strncmp(argv[ii],"-fseq",5)==0    && strlen(argv[ii])==5) ||
             (strncmp(argv[ii],"-fout",5)==0    && strlen(argv[ii])==5) ||
             (strncmp(argv[ii],"-fpwm",5)==0    && strlen(argv[ii])==5) ||
             (strncmp(argv[ii],"-nbps",5)==0    && strlen(argv[ii])==5) ||
             (strncmp(argv[ii],"-minN",5)==0    && strlen(argv[ii])==5) ||
             (strncmp(argv[ii],"-maxw3",6)==0   && strlen(argv[ii])==6) ||
             (strncmp(argv[ii],"-maxw4",6)==0   && strlen(argv[ii])==6) ||
             (strncmp(argv[ii],"-maxw5",6)==0   && strlen(argv[ii])==6) ||
             (strncmp(argv[ii],"-maxw6",6)==0   && strlen(argv[ii])==6) ||
             (strncmp(argv[ii],"-score",6)==0   && strlen(argv[ii])==6) ||
             (strncmp(argv[ii],"-mingap",7)==0  && strlen(argv[ii])==7) ||
             (strncmp(argv[ii],"-maxgap",7)==0  && strlen(argv[ii])==7) ||
             (strncmp(argv[ii],"-fracEM",7)==0  && strlen(argv[ii])==7) ||
             (strncmp(argv[ii],"-window",7)==0  && strlen(argv[ii])==7) ||
             (strncmp(argv[ii],"-fbackg",7)==0  && strlen(argv[ii])==7) ||
             (strncmp(argv[ii],"-extTrim",8)==0 && strlen(argv[ii])==8) ||
             (strncmp(argv[ii],"-verbose",8)==0 && strlen(argv[ii])==8)
            ) { goodArgument=1; }
         if (!goodArgument) { printf("argument: %s unknown\n",argv[ii]); exit(0);  }
      }
   }

   for (ii=1; ii<argc; ii++) {
      if (strncmp(argv[ii],"-fseq",5)==0 && argv[ii+1]!=NULL) {
         strcpy(mFileName,argv[ii+1]);
         seq=read_seq(&numSeq,seqLen,geneID,MAX_NUM_SEQ,MAX_SEQ_LENGTH,ChIPScore,argv[ii+1]);
         aveSeqLen=0; for (i=0; i<numSeq; i++) aveSeqLen +=seqLen[i]; aveSeqLen /=(double)numSeq;
         printf("\nInput data: number of sequences: %d, average length: %5.1f\n",numSeq,aveSeqLen);
         maxSeqLen=0; 
         for (i=0; i<numSeq; i++) {
            if (seqLen[i]>maxSeqLen) maxSeqLen=seqLen[i]; 
         }
 
         rseq=alloc_char_char(numSeq,maxSeqLen+1);
         oseq=alloc_char_char(numSeq,maxSeqLen+1);
         orseq=alloc_char_char(numSeq,maxSeqLen+1);

         reverse_seq(seq,rseq,numSeq,seqLen);
         // make a copy of the original sequences both strands
         for (i=0; i<numSeq; i++) {
            for (j=0; j<seqLen[i]; j++) { oseq[i][j]=seq[i][j]; orseq[i][j]=rseq[i][j]; } 
            oseq[i][seqLen[i]]='\0'; orseq[i][seqLen[i]]='\0'; 
         }
         //for (i=0; i<numSeq; i++) { printf("%s\n",seq[i]);  printf("%s\n",oseq[i]); }
         //for (i=0; i<numSeq; i++) { printf("%s\n",rseq[i]); printf("%s\n",orseq[i]);}
      }
      else if (strncmp(argv[ii],"-fbackg",7)==0 && argv[ii+1]!=NULL) {
         fp=fopen(argv[ii+1],"r");
         if (!fp) { perror(argv[ii+1]); exit(0); }
         bfreq=alloc_double(4);
         fscanf(fp,"%lf%lf%lf%lf",&bfreq[0],&bfreq[1],&bfreq[2],&bfreq[3]);
         fclose(fp); 
         sum=0; for (i=0; i<4; i++) sum +=bfreq[i];
         if (sum>0.0001) for (i=0; i<4; i++) bfreq[i]/=sum;
         else {
            printf("Error: reading background frequencies. see %s\n",argv[ii+1]); exit(0); 
         } 
         useAlternativeBackg=1;
      }
      else if (strncmp(argv[ii],"-pv",3)==0      && strlen(argv[ii])==3 && argv[ii+1]!=NULL) pvalueCutoff=atof(argv[ii+1]); 
      else if (strncmp(argv[ii],"-em",3)==0      && strlen(argv[ii])==3 && argv[ii+1]!=NULL) numEM=atoi(argv[ii+1]);
      else if (strncmp(argv[ii],"-ev",3)==0      && strlen(argv[ii])==3 && argv[ii+1]!=NULL) E_valueCutoff=atof(argv[ii+1]);
      else if (strncmp(argv[ii],"-pop",4)==0     && strlen(argv[ii])==4 && argv[ii+1]!=NULL) populationSize=atoi(argv[ii+1]); 
      else if (strncmp(argv[ii],"-gen",4)==0     && strlen(argv[ii])==4 && argv[ii+1]!=NULL) numGeneration=atoi(argv[ii+1]); 
      else if (strncmp(argv[ii],"-fout",5)==0    && strlen(argv[ii])==5 && argv[ii+1]!=NULL) strcpy(oFileName,argv[ii+1]); 
      else if (strncmp(argv[ii],"-fpwm",5)==0    && strlen(argv[ii])==5 && argv[ii+1]!=NULL) strcpy(pwmFileName,argv[ii+1]); 
      else if (strncmp(argv[ii],"-nbps",5)==0    && strlen(argv[ii])==5 && argv[ii+1]!=NULL) numBasePerSite=atoi(argv[ii+1]);
      else if (strncmp(argv[ii],"-minN",5)==0    && strlen(argv[ii])==5 && argv[ii+1]!=NULL) minsites=atoi(argv[ii+1]);
      else if (strncmp(argv[ii],"-maxw3",6)==0   && strlen(argv[ii])==6 && argv[ii+1]!=NULL) numTop3mer=atoi(argv[ii+1]);
      else if (strncmp(argv[ii],"-maxw4",6)==0   && strlen(argv[ii])==6 && argv[ii+1]!=NULL) numTop4mer=atoi(argv[ii+1]);
      else if (strncmp(argv[ii],"-maxw5",6)==0   && strlen(argv[ii])==6 && argv[ii+1]!=NULL) numTop5mer=atoi(argv[ii+1]);
      else if (strncmp(argv[ii],"-maxw6",6)==0   && strlen(argv[ii])==6 && argv[ii+1]!=NULL) numTop6mer=atoi(argv[ii+1]);
      else if (strncmp(argv[ii],"-score",6)==0   && strlen(argv[ii])==6 && argv[ii+1]!=NULL) useChIPscore=atoi(argv[ii+1]);
      else if (strncmp(argv[ii],"-fracEM",7)==0  && strlen(argv[ii])==7 && argv[ii+1]!=NULL) percentSeqEM=atof(argv[ii+1]);
      else if (strncmp(argv[ii],"-mingap",7)==0  && strlen(argv[ii])==7 && argv[ii+1]!=NULL) minSpaceWidth=atoi(argv[ii+1]);
      else if (strncmp(argv[ii],"-maxgap",7)==0  && strlen(argv[ii])==7 && argv[ii+1]!=NULL) maxSpaceWidth=atoi(argv[ii+1]);
      else if (strncmp(argv[ii],"-window",7)==0  && strlen(argv[ii])==7 && argv[ii+1]!=NULL) window=atoi(argv[ii+1]);
      else if (strncmp(argv[ii],"-extTrim",8)==0 && strlen(argv[ii])==8 && argv[ii+1]!=NULL) extTrim=atoi(argv[ii+1]);
      else if (strncmp(argv[ii],"-verbose",8)==0 && strlen(argv[ii])==8 && argv[ii+1]!=NULL) verbose=atoi(argv[ii+1]);
      else { }
   }

   // check for input parameters
   if (numGeneration<1)  { printf("\nError: numbe of generaton < 1\n"); exit(0); }
   if (populationSize<1) { printf("\nError: population size < 1\n");    exit(0); }
   if (minSpaceWidth<0)  { 
      printf("\nError: minimal number of unspecified bases in spaced dyads <0\n"); 
      printf("   check -mingap setting\n\n"); exit(0);
   }
   if (maxSpaceWidth<0)  { 
      printf("\nError: maximal number of unspecified bases in spaced dyads <0\n"); 
      printf("   check -maxgap setting\n\n"); exit(0);
   }
   if (minSpaceWidth>maxSpaceWidth) {
      printf("\nError: mingap setting must <= to maxgap setting\n\n"); exit(0); 
   }
   if (maxSpaceWidth+12>MAX_PWM_LENGTH) {
      printf("\nError: maxgap setting plus word lengths exceed <MAX_PWM_LENGTH>\n");
      printf("   For very long motifs, please set <MAX_PWM_LENGTH> in 'defines.h' accordingly\n\n"); exit(0); 
   }
   if (numEM<0) {
      printf("\nError: number of EM steps is zero\n"); exit(0); 
   }
   if (numEM==0) {
      printf("\nNote: numbre of EM steps = 0, no EM optimization is carried out\n");  
   }

   if (!useAlternativeBackg) {
      // base frequency from the whole data, both plus and reverse strands
      bfreq=base_frequency(numSeq,seq,seqLen);
      printf("\n[a,c,g,t] frequencies estimated in input data: %10.8f %10.8f %10.8f %10.8f\n\n",bfreq[0],bfreq[1],bfreq[2],bfreq[3]);
   }
   else {
      printf("\n[a,c,g,t] frequencies specified in backg: %10.8f %10.8f %10.8f %10.8f\n\n",bfreq[0],bfreq[1],bfreq[2],bfreq[3]);
   }

   maxWordSize=0;
   if (numTop3mer>maxWordSize) maxWordSize=numTop3mer;
   if (numTop4mer>maxWordSize) maxWordSize=numTop4mer;
   if (numTop5mer>maxWordSize) maxWordSize=numTop5mer;
   if (numTop6mer>maxWordSize) maxWordSize=numTop6mer;

   // any one, two or three: tetramer, pentamer, hexamer
   if (numTop3mer==0 && numTop4mer==0 && numTop5mer==0 && numTop6mer==0) {
      printf("\nError: maxw3, maxw4, maxw5, max6 all zero - no words for spaced dyads\n"); exit(0);
   }

   // number of sequences for EM
   if (percentSeqEM>1.0 || percentSeqEM<=0.0) { 
      printf("\nError: the fraction of sequences subject to EM is %3.2f\n",percentSeqEM); exit(0); 
   } 
   numSeqEM=(int)(percentSeqEM*numSeq);

   // memory callocations
   Iseq  =alloc_char(numSeq+1); 
   dyad  =alloc_chrs(populationSize,4);
   pwm   =alloc_double_double_double(populationSize,MAX_PWM_LENGTH,4);
   opwm  =alloc_double_double_double(populationSize,MAX_PWM_LENGTH,4);
   opwm2 =alloc_double_double(MAX_PWM_LENGTH,4);
   t1pwm =alloc_double_double(MAX_PWM_LENGTH,4);
   t2pwm =alloc_double_double(MAX_PWM_LENGTH,4);
   ipwm  =alloc_int_int_int(populationSize,MAX_PWM_LENGTH,4);
   pwmLen=alloc_int(populationSize);
   logpwm=alloc_double_double(MAX_PWM_LENGTH,4);
   score =alloc_double_double(numSeq,maxSeqLen);
   rscore=alloc_double_double(numSeq,maxSeqLen);
   sdyad =alloc_char_char(populationSize,MAX_PWM_LENGTH+1);
   word  =alloc_word(numWordGroup,maxWordSize);
   wheel =alloc_wheel(populationSize);
   siteEM=alloc_site(MAX_SITES);
   fitness=alloc_fitness(populationSize);
   emSeqLen=alloc_int(numSeqEM);
   pwmConsensus=alloc_char_char(populationSize,MAX_PWM_LENGTH+1);
   motifUniq=alloc_char(populationSize+1);
   scoreCutoff=alloc_int(populationSize);
   scoreDist=(Mfunction *)calloc(MAX_DIMENSION,sizeof(Mfunction));
   if (!scoreDist) { printf("calloc failed for scoreDist\n"); exit(0); }

   // if minN not specified, set the defaults accordingly
   if (minsites==-1) {
      // require at least one binding site per every 5000 bp
      minsites  =(int)(aveSeqLen*numSeq/numBasePerSite); 
   }
   minsitesEM=(int)(percentSeqEM*minsites);
   maxp=numSeqEM;

   seed=time(0);
   //seed=123456789;
   sgenrand(seed);

   // determine the distribution and critical cut point
   pwmDistCutoff=vector_similarity();

   // select a subset of sequences for EM only
   numSeqWithQualityScore=0; 
   for (i=0; i<numSeq; i++) {
      if (ChIPScore[i]>0) numSeqWithQualityScore++; 
   }
    
   tmpScore=alloc_double(numSeq);
   for (i=0; i<numSeq; i++) tmpScore[i]=ChIPScore[i];
   sort_double(tmpScore,numSeq);

   ChIPscoreCutoff=tmpScore[(int)(percentSeqEM*numSeq)];

   if (useChIPscore==1) {
      if (numSeqWithQualityScore<=(int)(percentSeqEM*numSeq)) {
         for (i=0; i<numSeq; i++) Iseq[i]='0';
         numSeqEMtmp1=0;
         for (i=0; i<numSeq; i++) {
            if (ChIPScore[i]>0) { 
               Iseq[i]='1'; numSeqEMtmp1++; 
            }
         }
         numSeqEMtmp2=0;
         for (i=0; i<numSeq; i++) {
            if (ChIPScore[i]<=0) { 
               Iseq[i]='1'; numSeqEMtmp2++; 
               if (numSeqEMtmp1+numSeqEMtmp2==numSeqEM) break;
            }
         }
      }
      else {
         for (i=0; i<numSeq; i++) Iseq[i]='0';
         numSeqEMtmp1=0; numSeqEMtmp2=0;
         for (i=0; i<numSeq; i++) {
            if (ChIPScore[i]>=ChIPscoreCutoff) { 
               Iseq[i]='1'; numSeqEMtmp1++; 
               if (numSeqEMtmp1==numSeqEM) break;
            }
         }
      }
   }
   else {
      useChIPscore=0;
      sample_without_replacement(Iseq,numSeqEM,numSeq);
   }
   if (tmpScore)  { free(tmpScore);  tmpScore=NULL;  }
   if (ChIPScore) { free(ChIPScore); ChIPScore=NULL; }
   // for (i=0; i<numSeq; i++) printf("%c",Iseq[i]); printf("\n");

   time(&t);
   fp=fopen("info.txt","w");
   fprintf(fp,"==========================================================================\n\n");
   fprintf(fp,"command line: ");
   for (i=0; i<argc; i++) fprintf(fp,"%s ",argv[i]); fprintf(fp,"\n\n");
   fprintf(fp,"maximal buffer length:\t\t\t\t\t%d\n",MAX_BUFFER_LENGTH);
   fprintf(fp,"maximal number of sequences allowed:\t\t\t%d\n",MAX_NUM_SEQ);
   fprintf(fp,"maximal number of bases per seq read:\t\t\t%d\n",MAX_SEQ_LENGTH);
   fprintf(fp,"maximal number of sites in a motif:\t\t\t%d\n",MAX_SITES);
   fprintf(fp,"input (ChIP) sequence file:\t\t\t\t%s\n",mFileName);
   fprintf(fp,"number of sequences in input file:\t\t\t%d\n",numSeq);
   fprintf(fp,"average sequence length:\t\t\t\t%d\n",(int)aveSeqLen);
   fprintf(fp,"total number of nucleotides:\t\t\t\t%d\n",(int)(aveSeqLen*numSeq));
   fprintf(fp,"number of sequences with ChIP quality score(>0):\t\t%d\n\n",numSeqWithQualityScore);
   fprintf(fp,"max number of generations:\t\t\t\t%d\n",numGeneration);
   fprintf(fp,"population size:\t\t\t\t\t%d\n",populationSize);
   fprintf(fp,"number of top-ranked 3-, 4-, 5-, and 6-mer:\t\t%d %d %d %d\n",numTop3mer,numTop4mer,numTop5mer,numTop6mer);
   fprintf(fp,"minimal & maximal spacer(d) in dyads\t\t\t%d %d\n",minSpaceWidth,maxSpaceWidth);
   if (useChIPscore && numSeqEMtmp2==0) fprintf(fp,"use seqs with quality scores >=%5.2f for EM to derive PWMs\n",ChIPscoreCutoff);
   else if (useChIPscore && numSeqEMtmp2!=0) 
       fprintf(fp,"use %d high scoring sequence and %d random seqs for EM to derive PWMs\n",numSeqEMtmp1,numSeqEMtmp2);
   else              fprintf(fp,"randomly select sequences for the EM algorithm\n");
   fprintf(fp,"fraction (number) input sequences subject to EM\t\t%3.2f (%d)\n",percentSeqEM,numSeqEM);
   fprintf(fp,"scale factor for converting (double)pwm to (int)pwm\t%4.0f\n",DOUBLE_TO_INT_SCALE);
   fprintf(fp,"number of EM steps:\t\t\t\t\t%d\n",numEM);
   fprintf(fp,"EM convergence criterion:\t\t\t\t\%e\n",PWM_CONVERGENCE);
   fprintf(fp,"maxp for sequences subject to EM:\t\t\t%d\n",maxp);
   fprintf(fp,"pwm score p-value cutoff for declaring binding site:\t%e\n",pvalueCutoff);
   fprintf(fp,"pseudo count:\t\t\t\t\t\t%5.4f\n",PSEUDO_COUNT);
   fprintf(fp,"minimal no. sites for each motif:\t\t\t%d\n",minsites);
   if (extTrim) fprintf(fp,"base extension and trimming? \t\t\t\tyes\n");
   else         fprintf(fp,"base extension and trimming? \t\t\t\tno\n");
   fprintf(fp,"sliding window size for comparing pwm similarity:\t%d\n",window);
   fprintf(fp,"log(e-value) cutoff:\t\t\t\t\t%5.2f\n",E_valueCutoff);
   fprintf(fp,"number of adjacent bases included in motif output:\t%d\n",FLANKING_BASES);
   fprintf(fp,"random seed:\t\t\t\t\t\t%ld\n\n",seed);
   fprintf(fp,"\njob started: %s\n", asctime(localtime(&t)));
   fprintf(fp,"=========================================================================\n\n");
   fflush(fp);

   printf("=====================================================================\n");
   printf("number of generations:\t\t\t\t\t\t%4d\n",numGeneration);
   printf("population size:\t\t\t\t\t\t%4d\n",populationSize);
   printf("PWM score p-value cutoff for binding site declaration:\t%e\n",pvalueCutoff);
   printf("number of top-ranked 3,4,5,6-mers for spaced dyads:\t%d %d %d %d\n",numTop3mer,numTop4mer,numTop5mer,numTop6mer);
   printf("minimal & maximal spacers for spaced dyad                      %1d%4d\n\n",minSpaceWidth,maxSpaceWidth);
   printf("number (percentage) of sequences selected for EM:\t%6d(%4.1f\%)\n",numSeqEM,100.0*(double)numSeqEM/(double)numSeq);
   printf("number of sequences with ChIP quality score(>0):\t\t %d\n",numSeqWithQualityScore);
   if (useChIPscore && numSeqEMtmp2==0) printf("use seqs with quality scores >=%5.2f for EM to derive PWMs\n",ChIPscoreCutoff);
   else if (useChIPscore && numSeqEMtmp2!=0) 
       printf("use %d high scoring sequence and %d random seqs for EM to derive PWMs\n",numSeqEMtmp1,numSeqEMtmp2);
   else printf("randomly select sequences for the EM algorithm\n");
   printf("number of EM steps:\t\t\t\t\t\t%4d\n",numEM);
   printf("minimal no. sites considered for a motif:\t\t\t%4d\n",minsites);
   printf("=====================================================================\n\n");

   printf("\nstarting GADEM... this may take a few hours to complete\n");
   printf("type: ctrl/z, then, bg, to run it in background\n\n");

   // determine seq length by counting only [a,c,g,t], seqLen is used in E-value calculation 
   effect_seq_length(seq,numSeq,seqLen,Iseq,emSeqLen);

   // determine the distribution and critical cut point
   pwmDistCutoff=vector_similarity();

   fq=fopen(oFileName,"w");
   fpwm=fopen(pwmFileName,"w");
   motifCn=0; noMotifFound=0; numCycle=0; numCycleNoMotif=0;
   do {

      // identify top-ranked k-mers (k=3,4,5,6) for spaced dyads
      printf("\nGADEM cycle %2d: enumerate and count k-mers...   ",numCycle+1);
      numWordGroup=word_for_dyad(word,seq,rseq,numSeq,seqLen,bfreq,&numTop3mer,&numTop4mer,&numTop5mer,&numTop6mer);
      printf("done.\n");

      // generating a "population" of spaced dyads
      printf("\ninitializing GA...   ");
      initialisation(dyad,populationSize,numWordGroup,word,minSpaceWidth,maxSpaceWidth);
      printf("done.\n\n");

      generationNoMotif=0;
      for (jjj=0; jjj<numGeneration; jjj++) {

         // convert spaced dyads to letter probability matrix
         dyad_to_pwm(word,populationSize,dyad,pwm,pwmLen);

         for (ii=0; ii<populationSize; ii++) {

            // to see from which spaced dyad a motif is derived
            pwm_profile(pwm[ii],pwmLen[ii],sdyad[ii]);
            // make a copy and then subject the copy to EM
            copy_pwm(pwm[ii],t1pwm,pwmLen[ii]); 
            // standarize pwm
            standardize_pwm(t1pwm,pwmLen[ii]);
            // for (j=0; j<4; j++) { for (m=0; m<pwmLen[0]; m++) printf("%4.2f ",t1pwm[m][j]); printf("\n"); } printf("\n\n"); 

            // EM on randomly selected sequences
            for (jj=0; jj<numEM; jj++) {
               log_pwm(t1pwm,logpwm,pwmLen[ii]);
               score_subsequences(numSeq,seq,rseq,seqLen,logpwm,pwmLen[ii],score,rscore,Iseq,bfreq);
               // compute p(zij|y=1) probability of binding sites started at seq i on position j
               normalize_global(score,rscore,seqLen,pwmLen[ii],numSeq,Iseq,maxp);
               // E-step
               construct_pwm(t2pwm,score,rscore,seq,rseq,seqLen,numSeq,pwmLen[ii],Iseq);
               standardize_pwm(t2pwm,pwmLen[ii]);
 
               pwmDiff=check_convergence(t1pwm,t2pwm,pwmLen[ii]);
               copy_pwm(t2pwm,t1pwm,pwmLen[ii]); 
               // for (j=0; j<4; j++) { for (m=0; m<pwmLen[0]; m++) printf("%4.2f ",t1pwm[m][j]); printf("\n"); } printf("\n\n"); 

               if (pwmDiff<=PWM_CONVERGENCE)  break;
            }

            // log-transform, convert (double)log(PWM) to an (int)PWM
            log_odds_to_int(t1pwm,ipwm[ii],pwmLen[ii],bfreq);
            // for (j=0; j<4; j++) { for (m=0; m<pwmLen[ii]; m++) printf("%5d ",ipwm[ii][m][j]); printf("\n"); } printf("\n\n"); 
            // compute score distribution of the (int)PWM using Staden's method 
            pDimension=pwm_score_dist(bfreq,ipwm[ii],pwmLen[ii],scoreDist);
            sort_ptable(scoreDist,pDimension);

            scoreCutoff[ii]=determine_cutoff(scoreDist,pDimension,pvalueCutoff);

            // declare binding sites using (int)PWM if a subseq scores at or above score cutoff
            nsitesEM=scan_sequence(siteEM,numSeq,seq,rseq,seqLen,ipwm[ii],pwmLen[ii],scoreCutoff[ii],Iseq,bfreq);
            // printf("maxp %d nsitesEM: %d pcutoff: %e scorecutoff: %d\n",maxp,nsitesEM,pvalueCutoff,scoreCutoff[ii]);
   
            // if more than user-specified minimal sites identified
            // since PWM is constructed on randomly selected seqs, reduce requirement
            if (nsitesEM>=max(2,(int)(0.4*minsitesEM))) {
               // construct pwm from the identified sites
               align_sites_count(siteEM,seq,rseq,nsitesEM,pwmLen[ii],opwm[ii]);
               standardize_pwm(opwm[ii],pwmLen[ii]);
               // compute E-value of the relative entroy score of the binding sites (motif)
               consensus_pwm(opwm[ii],pwmLen[ii],pwmConsensus[ii]);
               fitness[ii].value=E_value(opwm[ii],nsitesEM,bfreq,pwmLen[ii],numSeqEM,emSeqLen);
            }
            else {
               fitness[ii].value=DUMMY_FITNESS;
               for (i=0; i<pwmLen[ii]; i++) pwmConsensus[ii][i]='n'; pwmConsensus[ii][pwmLen[ii]]='\0'; 
            }
            fitness[ii].index=ii;

            if (verbose) { 
               printf("cyc.[%3d] gen.[%3d] pop.[%3d] spacedDyad: %s ",numCycle+1,jjj+1,ii+1,sdyad[ii]);
               for (j=strlen(sdyad[ii]); j<maxSpaceWidth+12; j++) printf(" ");
               printf("motifConsensus: %s",pwmConsensus[ii]);
               for (j=strlen(sdyad[ii]); j<maxSpaceWidth+12; j++) printf(" ");
               printf("fitness: %9.2f\n",fitness[ii].value);
            }
         }

         sort_fitness(fitness,populationSize);

         /*-----------------------------------------------------------------------------
         printf("generation %3d top %d:\n", jjj+1, populationSize);
         for (i=0; i<populationSize; i++) 
            printf("   pwmConsens: %s\tlogev: %6.2f\n",pwmConsensus[fitness[i].index],fitness[i].value);
         printf("\n");  
         -----------------------------------------------------------------------------*/
        
         numUniq=check_pwm_uniqueness_dist(opwm,pwmLen,populationSize,fitness,pwmDistCutoff,E_valueCutoff,motifUniq,window);

         printf("\nGADEM cycle[%3d] generation[%3d] number of unique motif: %d\n",numCycle+1,jjj+1,numUniq);
         fprintf(fp,"GADEM cycle[%3d] generation[%3d] number of unique motif(s): %d\n",numCycle+1,jjj+1,numUniq);
         for (i=0; i<populationSize; i++) {
            if (motifUniq[i]=='1') {
               printf("   spacedDyad: %s ",sdyad[fitness[i].index]);
               for (j=strlen(sdyad[fitness[i].index]); j<maxSpaceWidth+12; j++) printf(" ");
               printf("motifConsensus: %s ",pwmConsensus[fitness[i].index]);
               for (j=strlen(sdyad[fitness[i].index]); j<maxSpaceWidth+12; j++) printf(" ");
               printf(" fitness: %9.2f\n",fitness[i].value);

               fprintf(fp,"   spacedDyad: %s ",sdyad[fitness[i].index]);
               for (j=strlen(sdyad[fitness[i].index]); j<maxSpaceWidth+12; j++) fprintf(fp," ");
               fprintf(fp,"motifConsensus: %s ",pwmConsensus[fitness[i].index]);
               for (j=strlen(sdyad[fitness[i].index]); j<maxSpaceWidth+12; j++) fprintf(fp," ");
               fprintf(fp," fitness: %9.2f\n",fitness[i].value);
               //for (j=0; j<4; j++) {
               //   for (m=0; m<pwmLen[fitness[i].index]; m++) printf("%4.2f ",opwm[fitness[i].index][m][j]); printf("\n");
               //} printf("\n\n");
            }
         }
         printf("\n"); fprintf(fp,"\n"); fflush(fp);

         if (jjj<numGeneration-1) {

            // fitness based selection with replacement 
            roulett_wheel_fitness(fitness,populationSize,wheel);

            // mutation and crossover operations
            if (populationSize>1) {
               if (genrand()>=0.5) { mutation (dyad,numWordGroup,word,minSpaceWidth,maxSpaceWidth,wheel,populationSize,fitness,motifUniq); }
               else { crossover(dyad,numWordGroup,word,minSpaceWidth,maxSpaceWidth,wheel,populationSize,fitness,motifUniq); }
            }
            else { mutation (dyad,numWordGroup,word,minSpaceWidth,maxSpaceWidth,wheel,populationSize,fitness,motifUniq); }
         }
      }
      numCycle++;

      site  =alloc_site_site(numUniq+1,MAX_SITES);
      nsites=alloc_int(numUniq+1);
      pwmnewLen=alloc_int(numUniq+1);

      motifCn2=0; // motifCn per GADEM cycle
      for (ii=0; ii<populationSize; ii++) {

         id=fitness[ii].index;
   
         if (motifUniq[ii]=='0') continue;
 
         pDimension=pwm_score_dist(bfreq,ipwm[id],pwmLen[id],scoreDist);
         sort_ptable(scoreDist,pDimension);
         scoreCutoff[id]=determine_cutoff(scoreDist,pDimension,pvalueCutoff);
         // scan all sequences for sites using the EM-derived PWM 
         nsites[motifCn2]=scan_sequence_full(scoreDist,pDimension,site[motifCn2],numSeq,oseq,orseq,seqLen,ipwm[id],pwmLen[id],scoreCutoff[id],bfreq);

         sort_sites(site[motifCn2],nsites[motifCn2]);
 
         if (nsites[motifCn2]>=max(2,minsites)) {
            if (extTrim) {
               extend_alignment(site[motifCn2],numSeq,oseq,orseq,seqLen,nsites[motifCn2],pwmLen[id],&(pwmnewLen[motifCn2]));
            }
            else { pwmnewLen[motifCn2]=pwmLen[id]; } 

            align_sites_count(site[motifCn2],oseq,orseq,nsites[motifCn2],pwmnewLen[motifCn2],opwm2);
            standardize_pwm(opwm2,pwmnewLen[motifCn2]);

            logev=E_value(opwm2,nsites[motifCn2],bfreq,pwmnewLen[motifCn2],numSeq,seqLen);
            if (logev<=E_valueCutoff) {
               consensus_pwm(opwm2,pwmnewLen[motifCn2],pwmConsensus[id]);

               print_result_2(site[motifCn2],nsites[motifCn2],numSeq,oseq,orseq,seqLen,geneID,logev,opwm2,pwmnewLen[motifCn2],motifCn+1,sdyad[id],pwmConsensus[id],numCycle,pvalueCutoff,fq,fpwm);
               print_motif(site[motifCn2],nsites[motifCn2],oseq,orseq,seqLen,pwmnewLen[motifCn2],motifCn+1);
               // mask_sites(nsites[motifCn2],seq,rseq,seqLen,site[motifCn2],pwmnewLen[motifCn2]);
               motifCn++; motifCn2++; numCycleNoMotif=0;
            } 
         }
      }
      for (i=0; i<motifCn2; i++) {
         mask_sites(nsites[i],seq,rseq,seqLen,site[i],pwmnewLen[i]);
      }
      if (site[0])   { free(site[0]);   site[0]=NULL;   }
      if (site)      { free(site);      site=NULL;      }
      if (nsites)    { free(nsites);    nsites=NULL;    }
      if (pwmnewLen) { free(pwmnewLen); pwmnewLen=NULL; }
      
      if (motifCn2==0) numCycleNoMotif++;
    
      if (numCycleNoMotif==NUM_NO_MOTIF) noMotifFound=1;

   } while (!noMotifFound);
   fclose(fq); fclose(fpwm);

   time(&t);
   fprintf(fp,"\nfinished: %s\n", asctime(localtime(&t)));
   fclose(fp);

   if (dyad[0])         { free(dyad[0]);         dyad[0]=NULL;    }
   if (dyad)            { free(dyad);            dyad=NULL;       }
   if (seqLen)          { free(seqLen);          seqLen=NULL;     }
   if (pwm[0][0])       { free(pwm[0][0]);       pwm[0][0]=NULL;  }
   if (pwm[0])          { free(pwm[0]);          pwm[0]=NULL;     }
   if (pwm)             { free(pwm);             pwm=NULL;        }
   if (opwm2[0])        { free(opwm2[0]);        opwm2[0]=NULL;   }
   if (opwm2)           { free(opwm2);           opwm2=NULL;      }
   if (opwm[0][0])      { free(opwm[0][0]);      opwm[0][0]=NULL; }
   if (opwm[0])         { free(opwm[0]);         opwm[0]=NULL;    }
   if (opwm)            { free(opwm);            opwm=NULL;       }
   if (t1pwm[0])        { free(t1pwm[0]);        t1pwm[0]=NULL;   }
   if (t1pwm)           { free(t1pwm);           t1pwm=NULL;      }
   if (t2pwm[0])        { free(t2pwm[0]);        t2pwm[0]=NULL;   }
   if (t2pwm)           { free(t2pwm);           t2pwm=NULL;      }
   if (logpwm[0])       { free(logpwm[0]);       logpwm[0]=NULL;  }
   if (logpwm)          { free(logpwm);          logpwm=NULL;     }
   if (ipwm[0][0])      { free(ipwm[0][0]);      ipwm[0][0]=NULL; }
   if (ipwm[0])         { free(ipwm[0]);         ipwm[0]=NULL;    }
   if (ipwm)            { free(ipwm);            ipwm=NULL;       }
   if (pwmLen)          { free(pwmLen);          pwmLen=NULL;     }
   if (seq[0])          { free(seq[0]);          seq[0]=NULL;     }
   if (seq)             { free(seq);             seq=NULL;        }
   if (rseq[0])         { free(rseq[0]);         rseq[0]=NULL;    }
   if (rseq)            { free(rseq);            rseq=NULL;       }
   if (geneID[0])       { free(geneID[0]);       geneID[0]=NULL;  }
   if (oseq[0])         { free(oseq[0]);         oseq[0]=NULL;    }
   if (oseq)            { free(oseq);            oseq=NULL;       }
   if (orseq[0])        { free(orseq[0]);        orseq[0]=NULL;   }
   if (orseq)           { free(orseq);           orseq=NULL;      }
   if (geneID)          { free(geneID);          geneID=NULL;     }
   if (score[0])        { free(score[0]);        score[0]=NULL;   }
   if (score)           { free(score);           score=NULL;      }
   if (rscore[0])       { free(rscore[0]);       rscore[0]=NULL;  }
   if (rscore)          { free(rscore);          rscore=NULL;     }
   if (bfreq)           { free(bfreq);           bfreq=NULL;      }
   if (wheel)           { free(wheel);           wheel=NULL;      }
   if (fitness)         { free(fitness);         fitness=NULL;    }
   if (mFileName)       { free(mFileName);       mFileName=NULL;  }
   if (oFileName)       { free(oFileName);       oFileName=NULL;  }
   if (pwmFileName)     { free(pwmFileName);     pwmFileName=NULL;}
   if (sdyad[0])        { free(sdyad[0]);        sdyad[0]=NULL;   }
   if (sdyad)           { free(sdyad);           sdyad=NULL;      }
   if (siteEM)          { free(siteEM);          siteEM=NULL;     }
   if (pwmConsensus[0]) { free(pwmConsensus[0]); pwmConsensus[0]=NULL; }
   if (pwmConsensus)    { free(pwmConsensus);    pwmConsensus=NULL;    }
   if (word)       destroy_word(word,numWordGroup);

   return (1);
}

