#include "config.h"

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#ifdef HAVE_MALLOC_H
#include <malloc.h>
#endif

#include "ga_dyads.h"
#include "defines.h"
#include "random.h"

void mutation(Chrs **dyad,int numWordGroup,Words *word,int minSpaceWidth,int maxSpaceWidth,
   Wheel *wheel,int populationSize,Fitness *fitness,char *motifUniq) {

   register int i,j;
   int popuCn;                 // index of the number of dyads in population
   int whichToMutate;          // the spaced dyad for mutation
   int kmerGroup;              // pick a k-mer group (tetramer,pentamer,hexamer) in which a new word will be selected
   int whichPartDyad;          // pick a component of a spaced dyad for mutation
   int wID[3],wGroup[3]; // store temp wordID and groupID
   int found,spaceWidth,numIdentical;
   double rand;
   char lowEvalue;
   Chrs **tmpDyad;

   tmpDyad=alloc_chrs(populationSize,4);

   popuCn=0; 
   for (i=0; i<populationSize; i++) {
      if (motifUniq[i]=='1') {
         for (j=0; j<3; j++) {
            tmpDyad[popuCn][j].wordID   =dyad[fitness[i].index][j].wordID; 
            tmpDyad[popuCn][j].wordGroup=dyad[fitness[i].index][j].wordGroup; 
         }
         popuCn++;
      }
   }
   // printf("popuCn=%d\n",popuCn);

   numIdentical=0;
   while (popuCn<populationSize) {

      // sample with replacement with probability proportional to fitness score
      rand=(double)populationSize*genrand();
      whichToMutate=0; lowEvalue='0';
      for (j=0; j<populationSize; j++) {
         if (rand>=wheel[j].start && rand<=wheel[j].end) { 
            whichToMutate=wheel[j].index; lowEvalue=wheel[j].I_evalue; 
            // printf("mutation: %d fitness indicator: %c\n",whichToMutate,lowEvalue);
            break; 
         }
      }

      // if the spaced dyad did not produce a motif, likely all three components, w1,spacer,w2, are bad
      // it may be better off generating a new one instead.
      if (lowEvalue=='1') {
         kmerGroup=(int)(numWordGroup*genrand());
         if (kmerGroup==numWordGroup) kmerGroup--;
         rand=genrand();
         wID[0]=0;  // update right below
         for (j=0; j<word[kmerGroup].count; j++) {
            if (rand>=word[kmerGroup].prob_sta[j] && rand<word[kmerGroup].prob_end[j]) { wID[0]=j; break; }
         }
         wGroup[0]=kmerGroup;

         // choose a spacer width
         wGroup[1]=-1;  // dummy
         wID[1]=minSpaceWidth+(int)((maxSpaceWidth-minSpaceWidth+1)*genrand());
   
         // initialize w2 of dyad w1-spacer-w2 
         kmerGroup=(int)(numWordGroup*genrand());
         if (kmerGroup==numWordGroup) kmerGroup--;
         wID[2]=0; 
         rand=genrand();
         for (j=0; j<word[kmerGroup].count; j++) {
            if (rand>=word[kmerGroup].prob_sta[j] && rand<word[kmerGroup].prob_end[j]) { wID[2]=j; break; }
         }
         wGroup[2]=kmerGroup;
      }
      else { 
         for (j=0; j<3; j++) { 
            wID[j]   =dyad[whichToMutate][j].wordID;
            wGroup[j]=dyad[whichToMutate][j].wordGroup; 
         }
   
         // only one of the three blocks is allowed to be mutated
         // if spacer=0, only w1 and w2 are available
         if (maxSpaceWidth==0) { 
            whichPartDyad=(int)(2*genrand());
            if (whichPartDyad==1) whichPartDyad=2;
         }
         else {
            whichPartDyad=(int)(3*genrand());
            if (whichPartDyad==3) whichPartDyad--;
         }
         
         // if w1 of the dyad (w1-s-w2) is chosen
         if (whichPartDyad==0) {
   
            // decide which kmer kmerGroup (4-mer, 5-mer or 6-mer) from which w1 whill be replaced
            kmerGroup=(int)(numWordGroup*genrand());
            if (kmerGroup==numWordGroup) kmerGroup--;
   
            // select a k-mer from the k-group with prob proportional to z-score
            rand=genrand();
            wID[0]=0;
            for (j=0; j<word[kmerGroup].count; j++) {
               if (rand>=word[kmerGroup].prob_sta[j] && rand<=word[kmerGroup].prob_end[j]) { wID[0]=j; break; }
            }
            wGroup[0]=kmerGroup;        // mark the kmer kmerGroup from which w1 is selected
         }
   
         // if the spacer is chosen
         else if (whichPartDyad==1) {
            // choose a gap that is different from the current one
            do {
               spaceWidth=minSpaceWidth+(int)((maxSpaceWidth-minSpaceWidth+1)*genrand());
            } while (spaceWidth==wID[1]); 
            wID[1]=spaceWidth;
            wGroup[1]=-1;                 // dummy - not used
         }
   
         // if w2 of the dyad (w1-space-w2) is chosen 
         else {
            kmerGroup=(int)(numWordGroup*genrand());
            if (kmerGroup==numWordGroup) kmerGroup--;
   
            rand=genrand();
            wID[2]=0;
            for (j=0; j<word[kmerGroup].count; j++) {
               if (rand>=word[kmerGroup].prob_sta[j] && rand<word[kmerGroup].prob_end[j]) { wID[2]=j; break; }
            }
            wGroup[2]=kmerGroup;     // mark the kmer kmerGroup from which w2 is selected 
         }
      }

      // check if it is identical spaced dyads that exist
      found=0;
      for (i=0; i<popuCn; i++) {
         if (tmpDyad[i][0].wordID==wID[0] && tmpDyad[i][0].wordGroup==wGroup[0] &&  
             tmpDyad[i][1].wordID==wID[1] && tmpDyad[i][1].wordGroup==wGroup[1] &&  
             tmpDyad[i][2].wordID==wID[2] && tmpDyad[i][2].wordGroup==wGroup[2]) { found=1; break; }
      }
      if (!found) {
         for (j=0; j<3; j++) { 
            tmpDyad[popuCn][j].wordID   =wID[j]; 
            tmpDyad[popuCn][j].wordGroup=wGroup[j];
         }
         popuCn++;

      }
      else { numIdentical++; }
      // printf("popuCn=%d\n",popuCn);

      // avoiding endless while loop
      if (numIdentical==populationSize) break;
   }

   // update the population
   for (i=0; i<populationSize; i++) {
      for (j=0; j<3; j++) { 
         dyad[i][j].wordID   =tmpDyad[i][j].wordID; 
         dyad[i][j].wordGroup=tmpDyad[i][j].wordGroup; 
      } 
   }

   if (tmpDyad[0]) { free(tmpDyad[0]); tmpDyad[0]=NULL;  }
   if (tmpDyad)    { free(tmpDyad);    tmpDyad=NULL;     }
}
