#include "config.h"
#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#ifdef HAVE_MALLOC_H
#include <malloc.h>
#endif

#include "defines.h"

int range(double score);

/* normalize_per_seq is the subroutine for normalization (don't consider the sequences below the threshold) */
void normalize_global(double **score,double **rscore,int *seqLen,int motifLen,int numSeq,char *Iseq,int maxp){

   register int i,j,k,m;
   double totalOffset,localp,normalizeFactor,temp;
   int renormalize,totalSites;

   // convert log scale to original scale
   for(i=0; i<numSeq; i++){
      if (Iseq[i]=='0') continue;
      for(j=0; j<seqLen[i]-motifLen+1; j++){
         score[i][j]=exp(score[i][j]); 
         if (score[i][j]==1) score[i][j]=0.99999; 
         rscore[i][j]=exp(rscore[i][j]);
         if (rscore[i][j]==1) rscore[i][j]=0.99999; 
      }
   }
   totalSites=maxp;
   totalOffset=0; 
   for (i=0; i<numSeq; i++) {
      if (Iseq[i]=='0') continue;
      for (j=0; j<seqLen[i]-motifLen+1; j++) totalOffset += (score[i][j]+rscore[i][j]);
   }
   //printf("total=%5.3f\n",totalOffset);

   renormalize=1;
   while (renormalize) {
      renormalize=0;
      if (totalSites<1) break;

      // avoid dividing a small number by a small number
      if(totalOffset<10e-10 && totalSites>0){
         temp=range(totalOffset);
         normalizeFactor=(double)(totalOffset*pow(10,temp))/(double)(totalSites*pow(10,temp));
      }
      else {
         normalizeFactor=totalOffset/(double)totalSites;
      }

      totalOffset=0;
      for (i=0; i<numSeq; i++) {
         if (Iseq[i]=='0') continue;
         for (j=0; j<seqLen[i]-motifLen+1; j++) {
            if (score[i][j]<1) {
               score[i][j]/=normalizeFactor; 
               if (score[i][j]>=1) {
                  score[i][j]=1; totalSites--;
                  renormalize=1; 
               }
            }
            if (rscore[i][j]<1) {
               rscore[i][j]/=normalizeFactor; 
               if (rscore[i][j]>=1) {
                  rscore[i][j]=1; totalSites--;
                  renormalize=1; 
               }
            }
            if (score[i][j]<1)  totalOffset+= score[i][j]; 
            if (rscore[i][j]<1) totalOffset+=rscore[i][j]; 
         } 
      }
   };

  // smooth as in MEME
  for (i=0; i<numSeq; i++) {
      if (Iseq[i]=='0') continue;
      for(m=0; m<motifLen; m++) {
         for(j=0; j<(seqLen[i]-m)/motifLen; j++) {
            localp=0.0; 
            for(k=0; k<motifLen; k++) localp+=(score[i][j*motifLen+m+k]+rscore[i][j*motifLen+m+k]);
             
            if(localp>1.0) {
               for(k=0; k<motifLen; k++) {
                  score[i][j*motifLen+m+k] = score[i][j*motifLen+m+k]/localp;
                  rscore[i][j*motifLen+m+k]=rscore[i][j*motifLen+m+k]/localp;
               }
            }
         }
      }
   }
}

int range(double score) {

  int i;

  for(i=1; i<21; i++){
     if (score*pow(10,i)>=1.0) return (i);
  }
  printf("score is smaller than 10e-20\n");
  return(20);
}

void normalize_global_full(double **score,double **rscore,int *seqLen,int motifLen,int numSeq,int maxp){

   register int i,j,k,m;
   double totalOffset,localp,normalizeFactor,temp;
   int renormalize,totalSites;

   // convert log scale to original scale
   for(i=0; i<numSeq; i++){
      for(j=0; j<seqLen[i]-motifLen+1; j++){
         score[i][j]=exp(score[i][j]); 
         if (score[i][j]==1) score[i][j]=0.99999; 
         rscore[i][j]=exp(rscore[i][j]);
         if (rscore[i][j]==1) rscore[i][j]=0.99999; 
      }
   }
   totalSites=maxp;
   totalOffset=0; 
   for (i=0; i<numSeq; i++) {
      for (j=0; j<seqLen[i]-motifLen+1; j++) totalOffset += (score[i][j]+rscore[i][j]);
   }
   //printf("total=%5.3f\n",totalOffset);

   renormalize=1;
   while (renormalize) {
      renormalize=0;
      if (totalSites<1) break;

      // avoid dividing a small number by a small number
      if(totalOffset<10e-10 && totalSites>0){
         temp=range(totalOffset);
         normalizeFactor=(double)(totalOffset*pow(10,temp))/(double)(totalSites*pow(10,temp));
      }
      else {
         normalizeFactor=totalOffset/(double)totalSites;
      }

      totalOffset=0;
      for (i=0; i<numSeq; i++) {
         for (j=0; j<seqLen[i]-motifLen+1; j++) {
            if (score[i][j]<1) {
               score[i][j]/=normalizeFactor; 
               if (score[i][j]>=1) {
                  score[i][j]=1; totalSites--;
                  renormalize=1; 
               }
            }
            if (rscore[i][j]<1) {
               rscore[i][j]/=normalizeFactor; 
               if (rscore[i][j]>=1) {
                  rscore[i][j]=1; totalSites--;
                  renormalize=1; 
               }
            }
            if (score[i][j]<1)  totalOffset+= score[i][j]; 
            if (rscore[i][j]<1) totalOffset+=rscore[i][j]; 
         } 
      }
   };

  // smooth as in MEME
  for (i=0; i<numSeq; i++) {
      for(m=0; m<motifLen; m++) {
         for(j=0; j<(seqLen[i]-m)/motifLen; j++) {
            localp=0.0; 
            for(k=0; k<motifLen; k++) localp+=(score[i][j*motifLen+m+k]+rscore[i][j*motifLen+m+k]);
             
            if(localp>1.0) {
               for(k=0; k<motifLen; k++) {
                  score[i][j*motifLen+m+k] = score[i][j*motifLen+m+k]/localp;
                  rscore[i][j*motifLen+m+k]=rscore[i][j*motifLen+m+k]/localp;
               }
            }
         }
      }
   }
}
