#include "config.h"

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#ifdef HAVE_MALLOC_H
#include <malloc.h>
#endif

#include "ga_dyads.h"
#include "defines.h"

double find_pvalue(int score,Mfunction *M1,int dimension) {

    register int i;
    double p_result;

    p_result=0.0;
    for (i=0; i<dimension; i++) {
       if (M1[i].score>=score) p_result+=M1[i].pvalue;
       else break; 
    }
    return(p_result);
}

int ini_M(Mfunction *M0,double *bfreq,int **pwm,int col) {

   register int i,k,m;
   int sameElement;

   M0[0].score =pwm[col][0]; // score, etc - pwm element
   M0[0].pvalue=bfreq[0];    // polynomial coefficients - probabilities
 
   m=1;
   for (k=1; k<4; k++) {
      // add up the probabilities if scores/elements are the same
      sameElement=0;
      for (i=0; i<m; i++) {
         if (M0[i].score==pwm[col][k]) {
            M0[i].pvalue+=bfreq[k];
            sameElement=1;
            break; 
         }
      }
      // otherwise increase the dimension by one
      if (!sameElement) {
         M0[m].score =pwm[col][k];
         M0[m].pvalue=bfreq[k];
         m++;
         if (m==MAX_DIMENSION) { 
            printf("\nError: max p-table dimension reached\n"); 
            printf("  reset <MAX_DIMENSION> in defines.h\n");
            exit(0);
         }
      }
   }
   return (m);
}

int prod_M(Mfunction *M0,int dim0,Mfunction *M1,int dim1,Mfunction *M2) {

   register int i,j,k;
   int total,dim2;
   Mfunction *temp;

   temp=(Mfunction*)calloc(MAX_DIMENSION,sizeof(Mfunction)); 
   if (!temp) { printf("calloc failed for Mfunction temp in subroutine initial_p_table.c\n"); exit(0); }

   // add up the scores/elements 
   // times the probabilities
   total=0;
   for (i=0; i<dim0; i++) {
       for (j=0; j<dim1; j++) {
           temp[total].score =M0[i].score+M1[j].score;
           temp[total].pvalue=M0[i].pvalue*M1[j].pvalue;
           total++;
           if (total==MAX_DIMENSION) { 
              printf("\nError: max p-table dimension reached\n"); 
              printf("  reset <MAX_DIMENSION> in defines.h\n");
              exit(0);
           }
       }
   }
   
   // combine, if scores/elements are the same and probability is not zero
   for (k=0; k<total-1; k++) {
      if (temp[k].pvalue!=0.0) {
         for (j=k+1; j<total; j++) {
            if (temp[j].score==temp[k].score && temp[j].pvalue!=0.0) {
               temp[k].pvalue+=temp[j].pvalue;
               temp[j].pvalue=0.0;
            }
         }
      }
   }
   // update the list
   dim2=0;
   for (k=0; k<total; k++) {
      if (temp[k].pvalue>1e-15) {
         M2[dim2].score =temp[k].score;
         M2[dim2].pvalue=temp[k].pvalue;
         dim2++;
         if (dim2==MAX_DIMENSION) { 
            printf("\nError: max p-table dimension reached\n"); 
            printf("  reset <MAX_DIMENSION> in defines.h\n");
            exit(0);
         }
      }
   }
   if (temp)  {free(temp); temp=NULL; }

   return (dim2);
} 

int pwm_score_dist(double *bfreq,int **pwm,int pwmLen,Mfunction *scoreDist) {

   register int i,j;
   int dim0,dim1,dimension;
   Mfunction *M1,*M2;

   M1=(Mfunction* )calloc(4,sizeof(Mfunction));
   M2=(Mfunction* )calloc(MAX_DIMENSION,sizeof(Mfunction));
   if (!M1||!M2) { printf("calloc for M1,or M2 failed\n"); exit(0); }

   dim0=ini_M(scoreDist,bfreq,pwm,0); dimension=dim0;

   for (j=1; j<pwmLen; j++) {

      dim1=ini_M(M1,bfreq,pwm,j);
      dimension=prod_M(scoreDist,dim0,M1,dim1,M2);

      for (i=0; i<dimension; i++) {
         scoreDist[i].score =M2[i].score;
         scoreDist[i].pvalue=M2[i].pvalue;
      }
      dim0=dimension;
   }

   if (M1)  { free(M1); M1=NULL; }
   if (M2)  { free(M2); M2=NULL; }

   return (dimension);
}

int determine_cutoff(Mfunction *scoreDist,int dimension,double p_cutoff_ipwm) {

   register int j;
   int scoreCutoff;
   double pvalue;

   scoreCutoff=scoreDist[0].score;
   pvalue=scoreDist[0].pvalue;
   for (j=1; j<dimension; j++) {
      pvalue +=scoreDist[j].pvalue;
      if (pvalue>p_cutoff_ipwm) { scoreCutoff=scoreDist[j-1].score; break; }
   }

   return (scoreCutoff);
}

