#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "ga_dyads.h"
#include "defines.h"

int scan_sequence(Sites *site,int numSeq,char **seq,char **rseq,int *seqLen,int **ipwm,int pwmLen,
   int scoreCutoff,char *Iseq,double *bfreq) {

   register int ii,i,m;
   int plusScore,minusScore;
   int siteCn,found,posOnPlus;
   int *colAve;

   colAve=alloc_int(pwmLen);
   for (m=0; m<pwmLen; m++) {
      colAve[m]=0; for (i=0; i<4; i++) { colAve[m] +=(int)(ipwm[m][i]*bfreq[i]); } 
   }

   siteCn=0;
   for (ii=0; ii<numSeq; ii++) {
      if (Iseq[ii]=='0') continue;

      /* scan plus strand */
      for (i=0; i<seqLen[ii]-pwmLen+1; i++) {
         plusScore=0;
         for (m=0; m<pwmLen; m++) {
            switch(seq[ii][i+m]) {
               case 'a': plusScore +=ipwm[m][0]; break;
               case 'c': plusScore +=ipwm[m][1]; break;
               case 'g': plusScore +=ipwm[m][2]; break;
               case 't': plusScore +=ipwm[m][3]; break;
               default:  plusScore +=colAve[m];  break;
            }
         }

         posOnPlus=seqLen[ii]-i-pwmLen;
         minusScore=0;
         for (m=0; m<pwmLen; m++) {
            switch(rseq[ii][posOnPlus+m]) {
               case 'a': minusScore +=ipwm[m][0]; break;
               case 'c': minusScore +=ipwm[m][1]; break;
               case 'g': minusScore +=ipwm[m][2]; break;
               case 't': minusScore +=ipwm[m][3]; break;
               default:  minusScore +=colAve[m];  break;
            }
         }
         if (plusScore>=minusScore && plusScore>=scoreCutoff) {
            found=0;
            for (m=0; m<siteCn; m++) {
               // both site[m] and i are on plus strand
               if ((site[m].rev=='0' && abs(site[m].pos-i)<pwmLen && site[m].seq==ii) || 
               // i is on plus strand and site[m] on minus strand
                   (site[m].rev=='1' && abs(seqLen[ii]-site[m].pos-pwmLen-i)<pwmLen && site[m].seq==ii)) { found=1; break; }
            }
            if (!found) {
               //site[siteCn].score=plusScore;
               site[siteCn].pos=i;
               site[siteCn].seq=ii;
               site[siteCn].rev='0';
               siteCn++;
               if (siteCn==MAX_SITES) { 
                  printf("Error: max sites %d reached!\n",MAX_SITES); 
                  printf(" set <MAX_SITES> in defines.h accordingly\n");
                  exit(0); 
               }
            }
         }
         if (minusScore>plusScore && minusScore>=scoreCutoff) {
            found=0;
            for (m=0; m<siteCn; m++) {
               // both site[m] and i are on minus strand
               if ((site[m].rev=='0' && abs(site[m].pos-i)<pwmLen && site[m].seq==ii) || 
               // i is on minus strand and site[m] on plus strand
                   (site[m].rev=='1' && abs(seqLen[ii]-site[m].pos-pwmLen-i)<pwmLen && site[m].seq==ii)) { found=1; break; }
            }
            if (!found) {
               //site[siteCn].score=minusScore;
               site[siteCn].pos=posOnPlus;
               site[siteCn].seq=ii;
               site[siteCn].rev='1';
               siteCn++;
               if (siteCn==MAX_SITES) {
                  printf("Error: max sites %d reached!\n",MAX_SITES); 
                  printf(" set <MAX_SITES> in defines.h accordingly\n");
                  exit(0); 
               }
            }
         }
      }
   }
   //printf("total sites: %d\n",siteCn);

   if (colAve) { free(colAve); colAve=NULL; }

   return (siteCn);
}

int scan_sequence_full(Mfunction *scoreDist,int pDimension,Sites *site,int numSeq,char **seq,char **rseq,
   int *seqLen,int **ipwm,int pwmLen,int scoreCutoff,double *bfreq) {

   register int ii,i,m;
   int plusScore,minusScore;
   int siteCn,found,posOnPlus;
   int *colAve;

   colAve=alloc_int(pwmLen);
   for (m=0; m<pwmLen; m++) {
      colAve[m]=0; for (i=0; i<4; i++) { colAve[m] +=(int)(ipwm[m][i]*bfreq[i]); } 
   }

   siteCn=0;
   for (ii=0; ii<numSeq; ii++) {
      /* scan plus strand */
      for (i=0; i<seqLen[ii]-pwmLen+1; i++) {
         plusScore=0;
         for (m=0; m<pwmLen; m++) {
            switch(seq[ii][i+m]) {
               case 'a': plusScore +=ipwm[m][0]; break;
               case 'c': plusScore +=ipwm[m][1]; break;
               case 'g': plusScore +=ipwm[m][2]; break;
               case 't': plusScore +=ipwm[m][3]; break;
               default:  plusScore +=colAve[m];  break;
            }
         }

         posOnPlus=seqLen[ii]-i-pwmLen;
         minusScore=0;
         for (m=0; m<pwmLen; m++) {
            switch(rseq[ii][posOnPlus+m]) {
               case 'a': minusScore +=ipwm[m][0]; break;
               case 'c': minusScore +=ipwm[m][1]; break;
               case 'g': minusScore +=ipwm[m][2]; break;
               case 't': minusScore +=ipwm[m][3]; break;
               default:  minusScore +=colAve[m];  break;
            }
         }
         if (plusScore>=minusScore && plusScore>=scoreCutoff) {
            found=0;
            for (m=0; m<siteCn; m++) {
               // both site[m] and i are on plus strand
               if ((site[m].rev=='0' && abs(site[m].pos-i)<pwmLen && site[m].seq==ii) || 
               // i is on plus strand and site[m] on minus strand
                   (site[m].rev=='1' && abs(seqLen[ii]-site[m].pos-pwmLen-i)<pwmLen && site[m].seq==ii)) { found=1; break; }
            }
            if (!found) {
               //site[siteCn].score=plusScore;
               site[siteCn].pos=i;
               site[siteCn].seq=ii;
               site[siteCn].rev='0';
               site[siteCn].pvalue=find_pvalue(plusScore,scoreDist,pDimension);
               siteCn++;
               if (siteCn==MAX_SITES) {
                  printf("Error: max sites %d reached!\n",MAX_SITES); 
                  printf(" set <MAX_SITES> in defines.h accordingly\n");
                  exit(0); 
               }
            }
         }
         if (minusScore>plusScore && minusScore>=scoreCutoff) {
            found=0;
            for (m=0; m<siteCn; m++) {
               // both site[m] and i are on minus strand
               if ((site[m].rev=='0' && abs(site[m].pos-i)<pwmLen && site[m].seq==ii) || 
               // i is on minus strand and site[m] on plus strand
                   (site[m].rev=='1' && abs(seqLen[ii]-site[m].pos-pwmLen-i)<pwmLen && site[m].seq==ii)) { found=1; break; }
            }
            if (!found) {
               //site[siteCn].score=minusScore;
               site[siteCn].pos=posOnPlus;
               site[siteCn].seq=ii;
               site[siteCn].rev='1';
               site[siteCn].pvalue=find_pvalue(minusScore,scoreDist,pDimension);
               siteCn++;
               if (siteCn==MAX_SITES) {
                  printf("Error: max sites %d reached!\n",MAX_SITES); 
                  printf(" set <MAX_SITES> in defines.h accordingly\n");
                  exit(0); 
               }
            }
         }
      }
   }
   //printf("total sites: %d\n",siteCn);

   if (colAve) { free(colAve); colAve=NULL; }

   return (siteCn);
}
