#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "defines.h"

double **alloc_double_double(int ,int );

void standardize_pwm(double **pwm,int pwmLen) {

   register int i,j;
   double sum;

   for (i=0; i<pwmLen; i++) {
      sum=0.0; for (j=0; j<4; j++)  sum +=pwm[i][j]; 
     
      if (sum>PSEUDO_COUNT) { 
         for (j=0; j<4; j++) pwm[i][j] /=sum;  
      }
      else { 
         for (j=0; j<4; j++) pwm[i][j]  =0.25; 
      } 
   }
}

void log_pwm(double **pwm,double **logpwm,int pwmLen){

   register int i,j;

   for (i=0; i<pwmLen; i++) {
      if (pwm[i][0]<PSEUDO_COUNT || pwm[i][1]<PSEUDO_COUNT || pwm[i][2]<PSEUDO_COUNT || pwm[i][3]<PSEUDO_COUNT) {
          for (j=0; j<4; j++) logpwm[i][j]=log((pwm[i][j]+PSEUDO_COUNT)/(1.0+PSEUDO_COUNT*4));
      }
      else {
         for (j=0; j<4; j++) logpwm[i][j]=log(pwm[i][j]);
      }
   }
}

void log_odds_to_int(double **pwm,int **ipwm,int pwmLen,double *bfreq){

   register int i,j;
   double cellMin,cellMax;
   double **logodds;

   logodds=alloc_double_double(pwmLen,4);
 
   for (i=0; i<pwmLen; i++) {
      if (pwm[i][0]<PSEUDO_COUNT || pwm[i][1]<PSEUDO_COUNT || pwm[i][2]<PSEUDO_COUNT || pwm[i][3]<PSEUDO_COUNT) {
          for (j=0; j<4; j++) logodds[i][j]=(pwm[i][j]+PSEUDO_COUNT)/(1.0+PSEUDO_COUNT*4)/bfreq[j];
      }
      else {
          for (j=0; j<4; j++) logodds[i][j]=log(pwm[i][j]/bfreq[j]);
      }
   }

   cellMin=logodds[0][0]; cellMax=logodds[0][0]; 
   for (i=0; i<pwmLen; i++) {
      for (j=0; j<4; j++) {
        if (logodds[i][j]<cellMin) cellMin=logodds[i][j]; 
        if (logodds[i][j]>cellMax) cellMax=logodds[i][j]; 
      } 
   }

   if (cellMax-cellMin<0.01) {
      for (i=0; i<pwmLen;  i++) {
         for (j=0; j<4; j++) ipwm[i][j]=25;
      } 
   }
   for (i=0; i<pwmLen;  i++) {
      for (j=0; j<4; j++) {
         ipwm[i][j]=(int)(DOUBLE_TO_INT_SCALE*((logodds[i][j]-cellMin)/(cellMax-cellMin))); 
      } 
   }
   if (logodds[0]) { free(logodds[0]); logodds[0]=NULL; }
   if (logodds)    { free(logodds);    logodds=NULL;    }

   //for (j=0; j<4; j++) {
   //   for (i=0; i<pwmLen;  i++) printf("%4d",ipwm[i][j]); printf("\n");
   //}
}

